/*
**  Module    : control.c
**  Version   : 1.0
**  Date      : 
**  Authors   : Damiano Carra
**
**  Synopsis  : this module contains all the functions that deal with ...
**
**
**  History   : 19-JAN-2006 - Creation of the module.           
**
*/


# ifndef        STREAMING_H
#include "streaming.h"
# endif

/***************************
**  Function Definitions  **
***************************/

/*
**  Function  : void schedule(int, double, int, int, double, double)
**  Return    : none
**  Remarks   : this function inserts an event in the scheduled event list, 
**              setting all the event structure records 
**
*/
void schedule(int type, double time, Node *node, Node *from_node, int param)
{
  Event *ev;
  ev = new_event();
  ev->type      = type;
  ev->time      = time;
  ev->peer_id   = node;
  ev->from_peer = from_node;
  ev->param     = param;
  insert_event(ev);
  return;
}

/*
**  Function  : void start( int)
**  Return    : none
**  Remarks   : this function ... 
**                
**
*/
void start_tree(int num_stripes) 
{
int analyzed_nodes, node_index, neighbor_index, num_neighbors;
Node *neighbor_id, *node_id, *another_node_id;
double new_arrival_time;

   /* the server is fast, by definition */
   b_father = base_chunk_vect[1][num_deltas_pmf-1];  
  
   /* search a fast node that will be the server: if no fast nodes are found, pick randomly
      a node and change its bw */
   node_index = (int)uniform(0.0, (double)num_active_nodes, &seed);
   node_id = node_list[node_index];
   analyzed_nodes = 0;
   while((int)(bw_normalization*(*node_id).dwl_bw) != (int)b_father) {
            node_index = (node_index+1)%num_active_nodes; /* the modulo is used to manage the */
            node_id = node_list[node_index];           /* wrap of node list */
            analyzed_nodes++;                     
            if(analyzed_nodes > num_active_nodes)
               break;           
   }
   /* the selected node will be the server: all the settings should assure that the server
      will be never selected as children, neither as father */
   (*node_id).level               = 0;
   (*node_id).rate                = NUM_STRIPES*stripe_rate;
   (*node_id).end_chunk           = 0.0;
   (*node_id).num_children        = NUM_STRIPES;
   (*node_id).num_active_children = NUM_STRIPES;
   (*node_id).spare_dwl_bw        = 0.0;
   (*node_id).spare_upl_bw        = 0.0;
   (*node_id).leave_time          = 0.0; /* constant defined in math.h that means "infinity" */
   (*node_id).num_fathers         = -1;  
   (*node_id).stripe              = 0; 

   (*node_id).dwl_bw              = b_father/bw_normalization; //maybe it could be set to 0?
   (*node_id).upl_bw              = b_father/bw_normalization; 

   /* find among your neighbors 4 distinct nodes*/
   aux_index = 0;
   num_children = 0;
   num_neighbors = (*node_id).num_neighbors;
   while ((aux_index < num_neighbors)&&(num_children < NUM_STRIPES)) {
         /* check if the next neighbor in your list is available*/ 
         neighbor_id = (*node_id).neighbors[aux_index];      
         if ((*neighbor_id).level < 0) {
            /* if the neighbor is selected as a child, we swap it and put in first positions:
               in this way, the first num_children neighbors becomes my children*/
            if (aux_index > num_children){
                swap_ptr((*node_id).neighbors, num_children, aux_index);
            }
            num_children++;
            (*neighbor_id).level = 1; 
         }//if (neighbor is available)
         /* go to the next neighbor */
         aux_index++;              
   }//while(b_children_tot...
   
   #ifdef DEBUG
   if (num_children < NUM_STRIPES) printf("num_children < NUM_STRIPES\n");  
   #endif
   
   /* then, upload to num_stripes children, i.e. my first num_stripes neighbors */
   for (i = 0; i < num_children; i++){
         
         neighbor_id = (*node_id).neighbors[i];
   
         (*neighbor_id).level              = 1;
         (*neighbor_id).rate               = stripe_rate;
         (*neighbor_id).end_chunk          = 1/(stripe_threshold*stripe_rate);
         (*neighbor_id).spare_dwl_bw       = (*neighbor_id).dwl_bw - stripe_rate;  
         (*neighbor_id).num_fathers        = 1;  
         (*neighbor_id).num_active_fathers = 1;  
         (*neighbor_id).fathers_id[0]      = node_id;  
         (*neighbor_id).rate_of_each[0]    = stripe_rate;
         (*neighbor_id).delay_of_each[0]   = (*neighbor_id).end_chunk;
         (*neighbor_id).stripe_of_each[0]  = pow(2,i);
         /* the id of the stripe is divided into two halves. The first half registers 
            the original stripe, the other will be modified during the simulations */ 
         (*neighbor_id).stripe             = pow(2,i) + pow(2, half_size_int+i);
 
         /* update here the delay matrix */ 
         neighbor_index = (*neighbor_id).index; 
         stat_delay[DIFFUSION][neighbor_index] = (*neighbor_id).end_chunk;
         stat_delay[X_CONNECT][neighbor_index] = (*neighbor_id).end_chunk;
 
   } // for i
      
   for (i = 0; i < num_active_nodes; i++){
       another_node_id = node_list[i];
       if ((*another_node_id).level < 0) {
          new_arrival_time = 1/(stripe_threshold*stripe_rate) + 
                              negexp(10/(stripe_threshold*stripe_rate), &seed);
          schedule(JOIN, new_arrival_time, another_node_id, another_node_id, 0);
       }
   } // for i     
      
}

/*
**  Function  : void node_reached(int )
**  Return    : none
**  Remarks   : this function manages the necessary operation when
**              ....
*/
void node_reached(Node *node_id, Node *from_node) 
{
}

/*
**  Function  : void create_cross_connections(int )
**  Return    : none
**  Remarks   : this function manages the necessary operation when
**              ....
*/
void create_cross_connections(Node *node_id) 
{
}

/*
**  Function  : void drastic_search(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void delay_changed(Node *node_id, Node *from_node) 
{
}

/*
**  Function  : void reset()
**  Return    : none
**  Remarks   : this function ... 
**                
**
*/
void reset() 
{
int non_null_elem, counter, index;
int tmp_check1, tmp_check2, tmp_check3, stripe;
double rates_ratio, rounded_value, normalization_factor;


  /* heavy approach, that can be disabled with a great number of peers: 
     we analyze the node_list vector*/
  counter = 0;
  for (i=0; i < max_num_peers; i++){
         
      if ((*node_list[i]).end_chunk > 0) {

         stripe = (*node_list[i]).stripe >> half_size_int;
         stripe = (int)(log(stripe)/log(2)+0.5); 
         stat_dim_trees_sample[stripe]++;    
         
         stat_delay[DELT_POST][i] = stat_delay[X_CONNECT][i] - stat_delay[DIFFUSION][i];

         for (j=0; j < dim_stat_rates; j++){
             rounded_value = (double)((int)((*node_list[i]).rate*10000000))/10000000.0;  
             if (rounded_value <= stat_rates[KIND_RATE][j]) {
                stat_rates[DWNL_RATE][j]++;
                break;
             }
         }
         
         for (j=0; j < dim_stat_rates; j++){
             rates_ratio = ((*node_list[i]).upl_bw - (*node_list[i]).spare_upl_bw)/ 
                           ((*node_list[i]).dwl_bw - (*node_list[i]).spare_dwl_bw);

             if (rates_ratio <= stat_rates[KINDRATIO][j]) {
                stat_rates[UPL_DWL_R][j]++;
                break;
             }
         }    
         stat_topology[FATHERS][(*node_list[i]).num_fathers]++; 
         stat_topology[CHILDREN][(*node_list[i]).num_children]++;
         counter++; 
      
      
         for (k=0; k < N_PARAM_STAT_DELAY-1; k++){
              for (j=0; j < dim_delay_distrib; j++){
                   if ((stat_delay[k][i] <= delay_distrib[REF_DELAY][j])&&(stat_delay[k][i]!=0)){      
//                        delay_distrib[k][j]+= (1.0)/((double)num_active_nodes*max_iterations);
                          delay_collect[k][j]++;
                        break;
                      }
              }              
         }// for k
      
         k=N_PARAM_STAT_DELAY-1; 
         for (j=0; j < dim_delay_distrib; j++){
              if ((stat_delay[k][i] <= (delay_distrib[REF_DELAY][j]-delta_variation*((int)(1/(stripe_threshold*stripe_rate) + 0.5))))
                  &&(stat_delay[k][i] != -delta_variation*((int)(1/(stripe_threshold*stripe_rate) + 0.5)))) {      
//                  delay_distrib[k][j]+= (1.0)/((double)num_active_nodes*max_iterations);
                  delay_collect[k][j]++;
                  break;
              }
         }  
      } //if end_chunk > 0

  }
  
  if ((counter != num_active_nodes)){
     cumulative_arrived -= (num_active_nodes - counter);
     num_active_nodes = counter;
     }

  /* normalization of statistics on number of nodes per stripes */
  sort(stat_dim_trees_sample, 0, NUM_STRIPES -1);
  tmp_check1 = 0;
  for (j=0; j < NUM_STRIPES; j++){
       tmp_check1 += stat_dim_trees_sample[j];
  }//for j
  for (j=0; j < NUM_STRIPES; j++){
       stat_dim_trees[j] += stat_dim_trees_sample[j]/(tmp_check1*max_iterations);
       stat_dim_trees_sample[j] = 0;
  }//for j
    
  /* normalization of delay_distrib informations */
  normalization_factor = (double)((cumulative_arrived-cumulative_error)*(max_iterations));
  tmp_check1 = 0;
  tmp_check2 = 0;
  tmp_check3 = 0;
  for (j=0; j < dim_delay_distrib; j++){
       tmp_check1 += delay_collect[DIFFUSION][j];
       tmp_check2 += delay_collect[X_CONNECT][j];       
       tmp_check3 += delay_collect[DELT_POST][j];       
  }//for j
  for (j=0; j < dim_delay_distrib; j++){
       delay_distrib[DIFFUSION][j] += (double)delay_collect[DIFFUSION][j]/((double)(tmp_check1*max_iterations));
       delay_distrib[X_CONNECT][j] += (double)delay_collect[X_CONNECT][j]/((double)(tmp_check2*max_iterations));
       delay_distrib[DELT_POST][j] += (double)delay_collect[DELT_POST][j]/((double)(tmp_check3*max_iterations));
//       delay_distrib[DIFFUSION][j] += (double)delay_collect[DIFFUSION][j]/normalization_factor;
//       delay_distrib[X_CONNECT][j] += (double)delay_collect[X_CONNECT][j]/normalization_factor;
//       delay_distrib[DELT_POST][j] += (double)delay_collect[DELT_POST][j]/normalization_factor;
       delay_collect[DIFFUSION][j]    = 0.0;
       delay_collect[X_CONNECT][j]    = 0.0;
       delay_collect[DELT_POST][j]    = 0.0;
  }//for j

  /* statistics on number of peers */
     stat_num_peers[TOTAL___JOIN] += cumulative_arrived;
     stat_num_peers[TOTAL__LEAVE] += cumulative_left;
     stat_num_peers[PERCEN_LEAVE] += (double)cumulative_left/(double)cumulative_arrived;
     stat_num_peers[TOTAL_ACTIVE] += num_active_nodes;
     stat_num_peers[TOTAL__ERROR] += cumulative_error;
     stat_num_peers[PERCEN_ERROR] += (double)cumulative_error/(double)cumulative_arrived;
     stat_num_peers[MISSED__JOIN] += missed_join;
     stat_num_peers[ORPHAN_ACTIV] += cumulative_orphan;
     stat_num_peers[SOLVED__FAST] += cumulative_solved;

  /* normalization of topology informations */
  for (i=0; i < dim_stat_topology; i++){
      topology_distrib[FATHERS][i] += stat_topology[FATHERS][i]/normalization_factor;
      stat_topology[FATHERS][i] = 0.0;
      
      topology_distrib[CHILDREN][i] += stat_topology[CHILDREN][i]/normalization_factor;
      stat_topology[CHILDREN][i] = 0.0;  
  }
      
  /* normalization of rates informations */
  for (i=0; i < dim_stat_rates; i++){
      rates_distrib[DWNL_RATE][i] += stat_rates[DWNL_RATE][i]/(counter*max_iterations);
      stat_rates[DWNL_RATE][i] = 0.0;
      
      rates_distrib[UPL_DWL_R][i] += stat_rates[UPL_DWL_R][i]/(counter*max_iterations);
      stat_rates[UPL_DWL_R][i] = 0.0;
  }
  
  
  /* sort the stat_delay*/
  for (k=0; k < N_PARAM_STAT_DELAY; k++){
  
     sort(stat_delay[k], 0, max_num_peers - 1);
      /* find the firts non-null element (in one of the two matrices) */ 
     for (non_null_elem=0; non_null_elem < max_num_peers; non_null_elem++){
         if (stat_delay[k][non_null_elem] > 0.0000000001)
         break;
     }   
     /* update the samples */ 
     for (i=0; i < num_samples; i++){
         aux_index   = delay_samples[i][0] + non_null_elem;
         if (aux_index < max_num_peers) {
            delay_samples[i][3*k+1] += stat_delay[k][aux_index];
            delay_samples[i][3*k+2] += stat_delay[k][aux_index]*
                                       stat_delay[k][aux_index];
            delay_samples[i][3*k+3]++;
         }
     }//for i
  }// for k
  
  /* update temporal statistics on father */
  for (i=0; i < num_deltas_pmf; i++){
      /* expand the last variation up to the end */
      for (j = fathers_t_point[i]; j < MAX_TIME; j++){
           fathers_t_run[i][j] = fathers_t_run[i][fathers_t_point[i]];
          }
      /* make the mean */  
      for (j=0; j < MAX_TIME; j++){
           fathers_t_final[i][j] += fathers_t_run[i][j]/max_iterations;
           fathers_t_run[i][j]   = 0.0;
      }//for j
      fathers_t_index[i] = -1;
      fathers_t_point[i] = 0;
  }//for i     


  num_active_nodes = (int)(initial_percentage_active_nodes*actual_percentage_arriving_nodes*max_num_peers);
  num_non_active_nodes = max_num_peers - num_active_nodes;
  cumulative_arrived  = num_active_nodes;
  cumulative_left  = 0;
  cumulative_error = 0;
  missed_join = 0;
  cumulative_orphan = 0;
  cumulative_solved = 0;     
  leave_rate = 1.0/(p_churn*MAX_TIME);
  arrival_rate = (max_num_peers*actual_percentage_arriving_nodes*(1-initial_percentage_active_nodes))/time_span_for_arrival;
  arrival_rate += (max_num_peers*actual_percentage_arriving_nodes*initial_percentage_active_nodes)*leave_rate;
 
  /* re-initialize all the values for the next iteration */
  for (i=0; i < max_num_peers; i++){
         
      non_active_nodes[max_num_peers - 1 - i] = node_list[i];
      (*node_list[i]).level               = -1;
      (*node_list[i]).class               = -1;
      (*node_list[i]).rate                = 0.0;
      (*node_list[i]).end_chunk           = 0.0;
      (*node_list[i]).last_change         = 0.0;
      (*node_list[i]).num_children        = 0;
      (*node_list[i]).num_active_children = 0;
      if (i < num_active_nodes) {
           (*node_list[i]).leave_time   = HUGE_VAL;
           /* this is the case where "neighbors_id_basic" is used  
           (*node_list[i]).num_neighbors       = max_num_neighbors/2;
           for (j=0; j < max_num_neighbors/2; j++){
               index = (i + num_active_nodes + neighbors_id_basic[j])%num_active_nodes;
               (*node_list[i]).neighbors[j] = node_list[index];
           }//for j
           shuffle(max_num_neighbors/2, (*node_list[i]).neighbors, &seed);
            end of "neighbors_id_basic" case */   
           
           /* this is the case where "adj_matrix" is used  */
           j=0;
           while ((adj_matrix[i][j] >= 0)&&(j<max_num_neighbors/2)) {
           index = adj_matrix[i][j];
           (*node_list[i]).neighbors[j] = node_list[index];
           j++;
           }
           (*node_list[i]).num_neighbors = j;
           shuffle(j, (*node_list[i]).neighbors, &seed);
           /* end of "adj_matrix" case */   
      }
      else {
           (*node_list[i]).leave_time     = 0; 
           (*node_list[i]).num_neighbors  = 0;
      }   
      (*node_list[i]).num_fathers         = 0;  
      (*node_list[i]).num_active_fathers  = 0;  
      for (k=0; k < 2*NUM_STRIPES; k++){
          (*node_list[i]).fathers_id[k]     = NULL;
          (*node_list[i]).rate_of_each[k]   = 0.0;
          (*node_list[i]).stripe_of_each[k] = 0.0;
      }// for k
      (*node_list[i]).stripe       = 0;
      

      /* bandwidth assignation for active nodes */
      if (i < num_active_nodes) {
         rnd_value = uniform_0_1(&seed);
         extracted_dwl_bw = base_chunk_vect[1][0];
         extracted_upl_bw = base_chunk_vect[2][0];
         for (j=0; j<num_deltas_pmf-1; j++){
              if (rnd_value > base_chunk_vect[0][j]){
                 extracted_dwl_bw = base_chunk_vect[1][j+1];
                 extracted_upl_bw = base_chunk_vect[2][j+1];
              }//if
              else {break;}      
         }//for j
      
         (*node_list[i]).class = j;
         (*node_list[i]).dwl_bw = extracted_dwl_bw/bw_normalization;
         (*node_list[i]).upl_bw = extracted_upl_bw/bw_normalization; 
         (*node_list[i]).spare_dwl_bw = extracted_dwl_bw/bw_normalization;
         (*node_list[i]).spare_upl_bw = extracted_upl_bw/bw_normalization;
      }
          
   }//for i

   /* if, sooner or later, we start to use the module "statistics", here there are the updates 
      b_statistics(var_id[0],0);
      d_statistics(var_id[2],(double)queue);
      c_statistics(var_id[4], (double)numVosorg, current_time);
   */
}

/*
**  Function  : max_min_fairness_allocation()
**  Return    : change the values of b_allocated
**  Remarks   : 
*/        
void max_min_fairness_allocation(int num_children, double b_father, double *b_children_vect, 
                                   double *b_allocated, int *b_notcompl)
{
 double sum_spare;
 int    sum_notcompl, i, r;
 
     for (i=0; i<num_children; i++){
           b_allocated[i] = 0.0; 
           b_notcompl[i] = 1; 
      }//for i 
      sum_notcompl = num_children;
          while(sum_notcompl>0) {   
                sum_spare = 0;
                for (r=0; r<num_children; r++){
                    if(b_notcompl[r] == 1){    
                       b_allocated[r] += b_father/sum_notcompl;    
                       }//if
                    if(b_allocated[r] > b_children_vect[r]){
                       b_notcompl[r] = 0;
                       sum_spare += b_allocated[r] - b_children_vect[r];
                       b_allocated[r] = b_children_vect[r];    
                       }//if
                }//for r
                if(sum_spare > 0){
                   b_father = sum_spare;
                   sum_notcompl = 0;
                   for (r=0; r<num_children; r++){
                       sum_notcompl += b_notcompl[r];
                   }//for r
                }//if
                else { sum_notcompl = 0; }    
            }//while   
}

/*
**  Function  : void stop_simulation(void)
**  Return    : none
**  Remarks   : this funtion manages the end of the simulation; the output
**              file is generated with all the statistical results
**
*/
void stop_simulation(void)
{
 int i;
 double final_mean, final_variance, confidence_interval, unbias, final_num_nodes;
 double cumulative, cumulative2;
 
/* Here all the iterations are done and we start with the final statistics         */
strcpy(namefileout2, out_file_radix);
strcat(namefileout2, "_n.dat");
fout = fopen(namefileout2, "w"); 

end_CPU_time = clock(); 

final_num_nodes = stat_num_peers[TOTAL_ACTIVE]/max_iterations;


final_mean = 0;
final_variance = 0;
confidence_interval =0; 
unbias = 0;
fprintf(fout,"# CPU time (sec): %.2f\n", (double)(end_CPU_time - start_CPU_time)/CLOCKS_PER_SEC);
/*
fprintf(fout,"# format: #peers, delay_diffusion, c.i., delay_x_connect., c.i., ");
fprintf(fout,"# delat_delay_if_changed, c.i.\n");

for (i=0; i < num_samples; i++){

  if (delay_samples[i][0] < final_num_nodes){
    fprintf(fout,"%f",delay_samples[i][0]/final_num_nodes);

    for (k=0; k < N_PARAM_STAT_DELAY; k++){
        final_mean = delay_samples[i][3*k+1]/delay_samples[i][3*k+3];
        unbias = delay_samples[i][3*k+3]/(delay_samples[i][3*k+3]-1);
        final_variance = unbias*(delay_samples[i][3*k+2]/delay_samples[i][3*k+3] - (final_mean*final_mean));
        confidence_interval = 2.6*sqrt(final_variance/delay_samples[i][3*k+3]);
        fprintf(fout," %f %f", final_mean, confidence_interval);
    }
    fprintf(fout,"\n");
  }  
}
*/ 

fprintf(fout,"# statistics on number of nodes per stripes\n");
for (i=0; i < NUM_STRIPES; i++){
    fprintf(fout,"%d %f\n", i, stat_dim_trees[i]);
}//for i


fprintf(fout,"\n");

fprintf(fout,"# topological information\n");
fprintf(fout,"# indegree/outdegree, percentage_indegree, percentage_outdegree \n");
for (i=0; i < dim_stat_topology; i++){
     fprintf(fout,"%d %f %f\n", i, topology_distrib[FATHERS][i], topology_distrib[CHILDREN][i]);
}     

fprintf(fout,"\n");

fprintf(fout,"# rates information\n");
fprintf(fout,"# rate, percentage_dwnl, ratios, percentage_ratio_dwnl_over_upl \n");
for (i=0; i < dim_stat_rates; i++){
     fprintf(fout,"%f %f %f %f\n", rates_distrib[KIND_RATE][i], rates_distrib[DWNL_RATE][i], 
                                     rates_distrib[KINDRATIO][i], rates_distrib[UPL_DWL_R][i]);
}     

fprintf(fout,"\n");
cumulative  = 0.0;
cumulative2 = 0.0;
fprintf(fout,"# delay distribution\n");
fprintf(fout,"# delay, percentage of nodes during diffusion, percentage during x_conn, perc. delta  \n");
for (i=0; i < dim_delay_distrib; i++){
     cumulative  += delay_distrib[DIFFUSION][i];
     cumulative2 += delay_distrib[X_CONNECT][i];
     if ((delay_distrib[DIFFUSION][i] !=0)||(delay_distrib[X_CONNECT][i] !=0)||(delay_distrib[DELT_POST][i] !=0))
     fprintf(fout,"%f %e %e %f %e %e %e\n", delay_distrib[REF_DELAY][i], delay_distrib[DIFFUSION][i], 
                                   delay_distrib[X_CONNECT][i], 
                                   delay_distrib[REF_DELAY][i]-delta_variation*((int)(1/(stripe_threshold*stripe_rate) + 0.5)),
                                   delay_distrib[DELT_POST][i], cumulative, cumulative2);
}     

fprintf(fout,"\n");
fprintf(fout,"# statistics on number of peers\n");
fprintf(fout,"# mean number of join  : %f\n", stat_num_peers[TOTAL___JOIN]/max_iterations);
fprintf(fout,"# mean number of leave : %f\n", stat_num_peers[TOTAL__LEAVE]/max_iterations);
fprintf(fout,"# mean number of error : %f\n", stat_num_peers[TOTAL__ERROR]/max_iterations);
fprintf(fout,"# mean percen of leave : %f\n", stat_num_peers[PERCEN_LEAVE]/max_iterations);
fprintf(fout,"# mean percen of error : %f\n", stat_num_peers[PERCEN_ERROR]/max_iterations);
fprintf(fout,"# mean number of active: %f\n", stat_num_peers[TOTAL_ACTIVE]/max_iterations);
fprintf(fout,"# mean #missed_join    : %f\n", stat_num_peers[MISSED__JOIN]/max_iterations);
fprintf(fout,"# mean orphan of active: %f\n", stat_num_peers[ORPHAN_ACTIV]/max_iterations);
fprintf(fout,"# mean percen fast res : %f\n", stat_num_peers[SOLVED__FAST]/stat_num_peers[ORPHAN_ACTIV]);
fprintf(fout,"# mean percen solved af: %f\n", (stat_num_peers[ORPHAN_ACTIV] - 
                                               stat_num_peers[SOLVED__FAST] - 
                                               stat_num_peers[TOTAL__ERROR] )/
                                                        stat_num_peers[ORPHAN_ACTIV]);
fprintf(fout,"# mean percen error    : %f\n", stat_num_peers[TOTAL__ERROR]/stat_num_peers[ORPHAN_ACTIV]);

fprintf(fout,"#\n");
fprintf(fout,"#\n");

fprintf(fout,"# time statistics on num father\n");
fprintf(fout,"# time class1 class2 ...  \n");
for (j=0; j < MAX_TIME; j++){
  fprintf(fout,"%d ", j);
  for (i=0; i < num_deltas_pmf; i++){
      fprintf(fout,"%f ", fathers_t_final[i][j]);
  }//for i
  fprintf(fout,"\n");
}//for j     


fclose(fout);

/**************************************************************************/
/* end of file output processing                                             */
/**************************************************************************/
 
 fprintf(fp_statistics, "\n\n--------- FINAL STATISTICAL REPORT -------\n\n");

 for(i = 0; i < N_STAT_VAR; i++)
       {
        report_statistics(fp_statistics, var_id[i], TRUE);
        if(transient_is_over(var_id[i]))
           report_statistics(fp_statistics, var_id[i], FALSE);
       }


 fclose(fp_statistics);

exit (0);
}

/*
**  Function  : void leave_network(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void leave_network(Node *node_id) {

double delay;
int n_fathers, n_children, n_active_children, i, j, k, node_index, num_neighbors, 
    num_neigh_neighbors, stripe, child_n_fathers, child_n_active_fathers;
Node *father_id, *child_id, *neighbor_id;  
Booltype found;

/* since multiple factors can schedule a leave network, check here if this event is superflous, i.e.
   the node has already left */

 if ((*node_id).end_chunk > 0){
      
      /* update the global variable that counts the number of active nodes */
      non_active_nodes[num_non_active_nodes] = node_id;
      num_non_active_nodes++;
      num_active_nodes--;
      cumulative_left++;
      if((*node_id).leave_time == ERROR){
         cumulative_error++;
      }
      else{
         stripe = (*node_id).stripe >> half_size_int;
         stripe = (int)(log(stripe)/log(2)+0.5); 
         stat_dim_trees_sample[stripe]++;    
         
         stat_topology[FATHERS][(*node_id).num_fathers]++; 
         stat_topology[CHILDREN][(*node_id).num_children]++;
         node_index = (*node_id).index;
         for (k=0; k < N_PARAM_STAT_DELAY-1; k++){
             for (j=0; j < dim_delay_distrib; j++) {
                  if (stat_delay[k][node_index] <= delay_distrib[REF_DELAY][j]){      
                             delay_collect[k][j]++;
                           break;
                  }
             }              
          }// for k
      
         k=N_PARAM_STAT_DELAY-1; 
         for (j=0; j < dim_delay_distrib; j++){
             if ((stat_delay[k][node_index] <= (delay_distrib[REF_DELAY][j]-delta_variation*((int)(1/(stripe_threshold*stripe_rate) + 0.5))))
                     &&(stat_delay[k][node_index] != -delta_variation*((int)(1/(stripe_threshold*stripe_rate) + 0.5)))) {      
                     delay_collect[k][j]++;
                     break;
             }
         }              
         #ifdef DEBUG
         if (stat_delay[X_CONNECT][node_index] != (*node_id).end_chunk) {
             printf("end_chunk: %f, x_connect: %f, iterat. %d time: %f node %d\n", 
                     (*node_id).end_chunk, stat_delay[X_CONNECT][node_index], iterations, current_time, (*node_id).index);
         }           
         #endif
      }
            
      /* tell to (all) peer's fathers you are leaving the network; 
         first : put the leaving nodes in the last position of the children list
         second: decrease the number of children and increase the spare upload bw */
      n_fathers = (*node_id).num_fathers;
      for (i=0; i < n_fathers; i++){
           father_id         =  (*node_id).fathers_id[i];
           n_children        =  (*father_id).num_children;
           n_active_children =  (*father_id).num_active_children;

           #ifdef DEBUG
           if (n_children < 1) {
               printf("n_children: %d, iterat. %d time: %f node %d\n", 
                      n_children, iterations, current_time, (*node_id).index);
           }           
           #endif

           found = FALSE;
           for (j=0; j < n_children; j++){
               child_id = (*father_id).neighbors[j];
               if (child_id == node_id){
                   found= TRUE;
                   break;   
               }
           }

           #ifdef DEBUG
           if ((found==FALSE)) {
               printf("found0: %d, iterat. %d time: %f node %d\n", 
                       found, iterations, current_time, (*node_id).index);
           }           
           #endif

           if (j < n_active_children) {  
               swap_ptr((*father_id).neighbors, j, n_active_children-1);
               swap_ptr((*father_id).neighbors, n_active_children-1, n_children-1);
               (*father_id).num_active_children--;
               (*father_id).spare_upl_bw += (*node_id).rate_of_each[i];
           }           
           else {swap_ptr((*father_id).neighbors, j, n_children-1);}
           (*father_id).num_children--;
      
           /* setting the rate to 0 is necessary so the node cannot be contacted 
              by other nodes looking for a new father */
           (*node_id).rate_of_each[i] = 0;
      }
      (*node_id).num_fathers         = 0;  
      (*node_id).num_active_fathers  = 0;  
      (*node_id).spare_dwl_bw = (*node_id).dwl_bw;
      /* set the delay to zero, i.e., as if never reached */
      (*node_id).end_chunk = 0;
      /* set the total rate to 0*/
      (*node_id).rate = 0;
             
      /* update here the delay matrix */
      node_index = (*node_id).index; 
      stat_delay[DIFFUSION][node_index] = (*node_id).end_chunk;
      stat_delay[X_CONNECT][node_index] = (*node_id).end_chunk;
      stat_delay[DELT_POST][node_index] = (*node_id).end_chunk;
      
      /* tell to peer's children that now they are orphans; before, update they fields */
      n_children =  (*node_id).num_children;
      n_active_children =  (*node_id).num_active_children;
      for (i=0; i < n_children; i++){
           child_id  = (*node_id).neighbors[i];
           child_n_fathers        = (*child_id).num_fathers;
           child_n_active_fathers = (*child_id).num_active_fathers;

           found = FALSE;      
           for (j=0; j < child_n_fathers; j++){
               if ((*child_id).fathers_id[j] == node_id){
                  found = TRUE;      
                  break;   
               }
           }
    
           #ifdef DEBUG
           if ((child_n_fathers < 1)||(found==FALSE)) {
                 printf("n_fathers_z: %d, found: %d, iterat. %d time: %f node %d\n",
                         n_fathers, found, iterations, current_time, (*node_id).index);
           }           
           if ( ((i < n_active_children)&&(j > child_n_active_fathers))||
                ((i > n_active_children)&&(j < child_n_active_fathers))   )    {
                 printf("mismatch active-standby, iterat. %d time: %f node %d, child %d\n",
                         iterations, current_time, (*node_id).index, (*child_id).index);
           }           
           #endif

           if (j < child_n_active_fathers) {  
               cumulative_orphan++;
               swap_ptr((*child_id).fathers_id, j, child_n_active_fathers-1);
               swap((*child_id).rate_of_each,   j, child_n_active_fathers-1);
               swap((*child_id).delay_of_each,  j, child_n_active_fathers-1);
               swap_int((*child_id).stripe_of_each,   j, child_n_active_fathers-1);
                      
               swap_ptr((*child_id).fathers_id, child_n_active_fathers-1, child_n_fathers-1);
               swap((*child_id).rate_of_each,   child_n_active_fathers-1, child_n_fathers-1);
               swap((*child_id).delay_of_each,  child_n_active_fathers-1, child_n_fathers-1);
               swap_int((*child_id).stripe_of_each,   child_n_active_fathers-1, child_n_fathers-1);
               (*child_id).num_active_fathers--;
           
               (*child_id).rate         -= (*child_id).rate_of_each[child_n_fathers-1];
               (*child_id).spare_dwl_bw += (*child_id).rate_of_each[child_n_fathers-1];
               (*child_id).stripe       -= ((*child_id).stripe_of_each[child_n_fathers-1] << half_size_int);
               (*child_id).stripe       -=  (*child_id).stripe_of_each[child_n_fathers-1];
           }           
           else {
               swap_ptr((*child_id).fathers_id, j, child_n_fathers-1);
               swap((*child_id).rate_of_each,   j, child_n_fathers-1);
               swap((*child_id).delay_of_each,  j, child_n_fathers-1);
               swap_int((*child_id).stripe_of_each,   j, child_n_fathers-1);
           
               (*child_id).stripe       -=  (*child_id).stripe_of_each[child_n_fathers-1];
           }
           (*child_id).num_fathers--;

           delay = 0.0;   
//           delay = negexp((1.0/streaming_rate), &seed);   
           schedule(ORPHAN, current_time  + delay, child_id, node_id, (*child_id).stripe_of_each[child_n_fathers-1]);
      }
      (*node_id).num_active_children = 0;
      (*node_id).spare_upl_bw = (*node_id).upl_bw; 
       
      /* finally, delete all the neighbor relationships (egdes of the overlay graph), starting
         from num_children (the first num_children relationship will be modified by "search_new_father"*/
      num_neighbors  = (*node_id).num_neighbors;
      for (i=0; i < num_neighbors; i++){
           neighbor_id         = (*node_id).neighbors[i];
           num_neigh_neighbors = (*neighbor_id).num_neighbors;

           #ifdef DEBUG
           if (num_neigh_neighbors < 1) {
               printf("num_neigh_neighbors: %d, iterat. %d time: %f node %d\n", 
                      num_neigh_neighbors, iterations, current_time, (*node_id).index);
           }           
           #endif

           found = FALSE;
           for (j=0; j < num_neigh_neighbors; j++){
               child_id = (*neighbor_id).neighbors[j];
               if (child_id == node_id){
                   found= TRUE;
                   break;   
               }
           }

           #ifdef DEBUG
           if ((found==FALSE)) {
               printf("found0a: %d, iterat. %d time: %f node %d\n", 
                       found, iterations, current_time, (*node_id).index);
           }           
           #endif

           swap_ptr((*neighbor_id).neighbors, j, num_neigh_neighbors-1);
           (*neighbor_id).num_neighbors--;
      }
      (*node_id).num_neighbors = 0;
      
 }//if ((*node_id).end_chunk > 0)             
}


/*
**  Function  : void re_attach(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void attach(Node *node_id) {

int  node_index, num_neighbors, node_num_neighbors, 
     n_children, n_active_children, n_fathers, available_stripe, node_stripe;
Node *neighbor_id, *ngh_id;
Booltype found;
double available_rate, disapper_time, delay_of_each;


   node_num_neighbors = (*node_id).num_neighbors; 
  
   /* attach to a main father; since the node selection was random, the neighbors
      can be considered randomly ordered */
   n_fathers = 0;   /* the node has no father... */                     
   node_stripe = 0; /* ... and no stripes        */
   
   if (attach_to_the_best){
   /* selects the best delay */
   }//if (attach_to_the_best)

   for (i=0; i < node_num_neighbors; i++){
        
        if (n_fathers == 1)
           break; 
            
        neighbor_id    = (*node_id).neighbors[i];
        available_rate =  MIN((*neighbor_id).rate, (*neighbor_id).spare_upl_bw);
        if ((*neighbor_id).end_chunk > current_time) {
             available_rate = 0;
        }   
        n_children     =  (*neighbor_id).num_children;
        available_stripe = (*neighbor_id).stripe >> half_size_int;
        available_stripe = node_stripe^(node_stripe|available_stripe);
         
        if ((available_rate >= stripe_rate) && (available_stripe > 0 ) &&
           ((*neighbor_id).leave_time > current_time) && (n_children < max_num_children))  {
            /* a father was found, attach to it 
               - first : search the position in the neighbor list
               - second: switch to the correct position 
               - third : update the father record
               - forth : update the node record 
            */
            found = FALSE;      
            num_neighbors = (*neighbor_id).num_neighbors;
            for (j=num_neighbors-1; j >= 0; j--){ //for efficiency, start at the end of the neighbor list
                ngh_id = (*neighbor_id).neighbors[j];
                if (ngh_id == node_id){
                    found = TRUE;      
                    break;
                }     
            }

            #ifdef DEBUG
            if ((found==FALSE)) {
                printf("found0b: %d, iterat. %d time: %f node %d\n", 
                        found, iterations, current_time, (*node_id).index);
            }           
            #endif
          
            n_children         =  (*neighbor_id).num_children;
            n_active_children  =  (*neighbor_id).num_active_children;
            swap_ptr((*neighbor_id).neighbors, j, n_children);
            swap_ptr((*neighbor_id).neighbors, n_active_children, n_children);
            (*neighbor_id).num_children++;
            (*neighbor_id).num_active_children++;
            (*neighbor_id).spare_upl_bw -= stripe_rate;    
         
            /* update the node */
            found = FALSE;      
            delay_of_each = 0;      
            available_stripe = choose_rnd_stripe(available_stripe);
            for (j=0; j < (*neighbor_id).num_active_fathers; j++){
                 if ((*neighbor_id).stripe_of_each[j] == available_stripe){
                      delay_of_each = (*neighbor_id).delay_of_each[j];
                      found = TRUE;      
                      break;
                 }
            }
            
            #ifdef DEBUG
            if ((found==FALSE)) {
                printf("found0c: %d, iterat. %d time: %f node %d\n", 
                        found, iterations, current_time, (*node_id).index);
            }           
            #endif
            
            (*node_id).level     = (*neighbor_id).level + 1;
            (*node_id).rate      = stripe_rate;  
            (*node_id).end_chunk = delay_of_each + 1/(stripe_threshold*(*node_id).rate);
            (*node_id).num_fathers++;
            (*node_id).num_active_fathers++;
            
            (*node_id).fathers_id[0]     = neighbor_id;  
            (*node_id).rate_of_each[0]   = stripe_rate;
            (*node_id).delay_of_each[0]  = (*node_id).end_chunk;
            (*node_id).stripe_of_each[0] = available_stripe;
            n_fathers++;
            (*node_id).stripe += (available_stripe << half_size_int);
            (*node_id).stripe += available_stripe;
                 
            (*node_id).spare_dwl_bw = (*node_id).dwl_bw - stripe_rate;
        }
   }

   /* if no fathers was found, try later: in particular, select a random time exp. distrib. with mean 10*T_chunk; 
      we must also set the end.chunk to a non-zero value */
   if ((*node_id).num_fathers == 0){
      (*node_id).end_chunk = 0.1;
      new_arrival_time = current_time + negexp(10/(stripe_threshold*stripe_rate), &seed);
      schedule(JOIN, new_arrival_time, node_id, node_id, 0);
   }
   
   else {
   
      /* schedule a disappear */ 
      disapper_time = current_time + negexp((1/leave_rate), &seed);
      schedule(LEAVE, disapper_time, node_id, node_id, 0);
      (*node_id).leave_time = disapper_time;  
              
      /* update here the delay matrix */ 
      node_index = (*node_id).index; 
      stat_delay[DIFFUSION][node_index] = (*node_id).end_chunk;
      stat_delay[X_CONNECT][node_index] = (*node_id).end_chunk;
   
      /* update here the delay over time */ 
      if (((*node_id).leave_time > MAX_TIME)&&(fathers_t_index[(*node_id).class] < 0)){
            fathers_t_index[(*node_id).class] = (*node_id).index;
            for (i = fathers_t_point[(*node_id).class]; i < current_time; i++){
                fathers_t_run[(*node_id).class][i] = 
                        fathers_t_run[(*node_id).class][fathers_t_point[(*node_id).class]];
            }
            fathers_t_run[(*node_id).class][(int)current_time] = (*node_id).num_fathers;
            fathers_t_point[(*node_id).class] = (int)current_time;
      }  
              
      /* launch the "update indegree" event*/
      if (current_time < time_span_for_arrival){
           new_arrival_time = current_time + negexp(10/(stripe_threshold*stripe_rate), &seed);
      }
      else new_arrival_time = current_time;
      new_arrival_time = current_time + negexp(10/(stripe_threshold*stripe_rate), &seed);
      if ((let_diffusion)&&(new_arrival_time < (*node_id).leave_time))
            schedule(CHECK_INDEGREE, new_arrival_time, node_id, node_id, 0);
    }
}


/*
**  Function  : void drastic_search(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void update_indegree(Node *node_id) {

int n_fathers, i, node_num_neighbors, node_stripe, available_stripe, n_children, num_neighbors, 
    n_active_fathers, n_active_children, node_index;
Node *father_id, *neighbor_id, *ngh_id;    
Booltype found;
double available_rate, total_rate, delay_of_each;


   n_fathers          = (*node_id).num_fathers;
   /* after the update, check your connectivity status and, if the number of fathers is too low, 
        try to find other fathers*/
   if (((*node_id).leave_time > 0)&&(n_fathers < threshold_for_fathers)&&(((*node_id).rate > 0))){
   
        node_num_neighbors = (*node_id).num_neighbors; 
        node_stripe = 0; /* register the stripes */
        for (i=0; i < n_fathers; i++){
           node_stripe += (*node_id).stripe_of_each[i];
        }

        for (i=0; i < node_num_neighbors; i++){
            neighbor_id    = (*node_id).neighbors[i];
            available_rate =  MIN((*neighbor_id).rate, (*neighbor_id).spare_upl_bw);
            if ((*neighbor_id).end_chunk > current_time) {
                available_rate = 0;
            }   
            n_children     =  (*neighbor_id).num_children;
            available_stripe = (*neighbor_id).stripe >> half_size_int;
            available_stripe = node_stripe^(node_stripe|available_stripe);

            /* check if it is already my father */
            for (j=0; j < n_fathers; j++){
                 father_id  =  (*node_id).fathers_id[j];
                 if (father_id == neighbor_id)
                     available_stripe = 0;
            }
                             
            if ((available_rate >= stripe_rate) && (available_stripe > 0 ) &&
               ((*neighbor_id).leave_time > current_time) && (n_children < max_num_children))  {
               /* a father was found, attach to it  */
               found = FALSE;      
               num_neighbors = (*neighbor_id).num_neighbors;
               for (j=num_neighbors-1; j >= 0; j--){ //for efficiency, start at the end of the neighbor list
                   ngh_id = (*neighbor_id).neighbors[j];
                   if (ngh_id == node_id){
                       found = TRUE;      
                       break;
                   }     
               }

               #ifdef DEBUG
               if ((found==FALSE)) {
                printf("found0e: %d, iterat. %d time: %f node %d\n", 
                        found, iterations, current_time, (*node_id).index);
               }           
               #endif
          
               swap_ptr((*neighbor_id).neighbors, j, n_children); /* n_children contains the correct num. of childr.*/
               (*neighbor_id).num_children++;
                 
               /* update the node */
               found = FALSE;      
               delay_of_each = 0;      
               available_stripe = choose_rnd_stripe(available_stripe);
               for (j=0; j < (*neighbor_id).num_active_fathers; j++){
                    if ((*neighbor_id).stripe_of_each[j] == available_stripe){
                         delay_of_each = (*neighbor_id).delay_of_each[j];
                         found = TRUE;      
                         break;
                    }
               }
               
               #ifdef DEBUG
               if ((found==FALSE)) {
                   printf("found0c1: %d, iterat. %d time: %f node %d\n", 
                           found, iterations, current_time, (*node_id).index);
               }           
               #endif

               (*node_id).fathers_id[n_fathers]     = neighbor_id;  
               (*node_id).rate_of_each[n_fathers]   = stripe_rate;
               (*node_id).delay_of_each[n_fathers]  = delay_of_each + 1/(stripe_threshold*stripe_rate);
 
               (*node_id).stripe_of_each[n_fathers] = available_stripe;
               (*node_id).num_fathers++;

               n_fathers++;
               (*node_id).stripe += available_stripe;
               node_stripe += available_stripe;
                 
               if ((*node_id).rate <= (stripe_threshold*stripe_rate*0.99)) {
                    /* mark as active father: swap its position from last place to last of active */     
                    n_active_fathers = (*node_id).num_active_fathers;
                    swap_ptr((*node_id).fathers_id, n_fathers - 1, n_active_fathers);
                    swap((*node_id).rate_of_each,   n_fathers - 1, n_active_fathers);
                    swap((*node_id).delay_of_each,  n_fathers - 1, n_active_fathers);
                    swap_int((*node_id).stripe_of_each,   n_fathers - 1, n_active_fathers);
                    (*node_id).num_active_fathers++;
                    n_active_fathers++;    
                    total_rate = 0;
                    for (j=0; j<n_active_fathers; j++){
                         total_rate += (*node_id).rate_of_each[j];
                    }
                   (*node_id).rate         = total_rate;  
                   (*node_id).spare_dwl_bw = (*node_id).dwl_bw - total_rate;
                   (*node_id).stripe += (available_stripe << half_size_int);
                   (*node_id).end_chunk = MAX((*node_id).end_chunk, (*node_id).delay_of_each[n_active_fathers-1]);
                 
                   n_active_children  =  (*neighbor_id).num_active_children;
                   swap_ptr((*neighbor_id).neighbors, n_active_children, n_children);
                   (*neighbor_id).num_active_children++;
                   /* rate offered to the node is the stripe rate */ 
                   (*neighbor_id).spare_upl_bw -= stripe_rate;    
               }
               /* a piece of optimization code: it avoids cannibalization of available fathers       
               if ((n_fathers >= threshold_for_fathers/2)&&(current_time<time_span_for_arrival))
                   break; */   
           }
       } //for (i=0; i < node_num_neighbors; i++)

       node_index = (*node_id).index;  
       stat_delay[X_CONNECT][node_index] = (*node_id).end_chunk;
              
       /* check if some of your non-active fathers has no spare upload bandwidth 
       n_fathers        = (*node_id).num_fathers;
       n_active_fathers = (*node_id).num_active_fathers;
       for (i=n_active_fathers; i < n_fathers; i++){
            father_id  =  (*node_id).fathers_id[i];
            if ( (*father_id).spare_upl_bw < stripe_rate){
                schedule(ORPHAN, current_time, node_id, father_id);
            }
       }
       */
       
      /* update temporal statistics for the number of fathers */
      if (fathers_t_index[(*node_id).class] == (*node_id).index) {
            for (i = fathers_t_point[(*node_id).class]; i < current_time; i++){
                fathers_t_run[(*node_id).class][i] = fathers_t_run[(*node_id).class][fathers_t_point[(*node_id).class]];
            }
            fathers_t_run[(*node_id).class][(int)current_time] = (*node_id).num_fathers;
            fathers_t_point[(*node_id).class] = (int)current_time;
      }  
   } //if (search_new_fathers) 

   /* launch the new "update indegree" event*/
   new_arrival_time = current_time + negexp(time_span_for_arrival, &seed);
   if (new_arrival_time < (*node_id).leave_time)
       schedule(CHECK_INDEGREE, new_arrival_time, node_id, node_id, 0);
       

}



/*
**  Function  : void generate_new_join(void);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void generate_new_join() {

Node *node_id, *contact_id;
Booltype found;
int  i, contact_index, neighbor_index, num_neighbors, node_num_neighbors, 
     visited_nodes;
//Node , *neighbor_id, *ngh_id;
//double available_rate, disapper_time, delay_of_each;


   found = FALSE;
   
   /* launch the next generation of node, if we can */
   if (current_time > time_span_for_arrival) { 
      arrival_rate = (max_num_peers*actual_percentage_arriving_nodes)*leave_rate;
   }
   else {
      arrival_rate = (max_num_peers*actual_percentage_arriving_nodes*(1-initial_percentage_active_nodes))/time_span_for_arrival;
      arrival_rate += (max_num_peers*actual_percentage_arriving_nodes*initial_percentage_active_nodes)*leave_rate;
      arrival_rate += (max_num_peers*actual_percentage_arriving_nodes*(1-initial_percentage_active_nodes))*(current_time/time_span_for_arrival)*leave_rate;
      
//      arrival_rate *= (1+current_time*leave_rate + 
//                        (initial_percentage_active_nodes/(1-initial_percentage_active_nodes))*time_span_for_arrival);
   }
   
   new_arrival_time = current_time + negexp((1/arrival_rate), &seed);
   schedule(GENERATION, new_arrival_time, NULL, NULL, 0);
      
   /* check if there are some structure available; if not, generate a bunch of new ones */
   if (num_non_active_nodes == 0) {
   
      missed_join++;
   }
   
   /* take the first available structure and launch the 'join' event (that will take care
      of generating its own 'leave' event) */

   else{
       
       num_non_active_nodes--;
       node_id = non_active_nodes[num_non_active_nodes];
       /* check if there are some events on the fly */
       found = TRUE;
       if ((*node_id).last_change > current_time){
          found = FALSE;
          i = 0;
          node_id = non_active_nodes[i];
          while (((*node_id).last_change > current_time)&&(i<num_non_active_nodes)){
                 i++;
                 node_id = non_active_nodes[i];
          }
          if (i<num_non_active_nodes) { 
              swap_ptr(non_active_nodes, i, num_non_active_nodes);
              found = TRUE;
          }
          else {num_non_active_nodes++;
                missed_join++;}
       }
   }
   
   
   if (found == TRUE) {
      /* update the global variable that counts the number of active nodes */
      num_active_nodes++;
      cumulative_arrived++;
   
      /* initialize the node's basic parameters: probably some of them are already set, so you can leave them out  */
      (*node_id).level               = -1;
      (*node_id).rate                = 0.0; //leave out?
      (*node_id).end_chunk           = 0.0;
      (*node_id).last_change         = 0.0; 
      (*node_id).num_children        = 0;   //leave out?
      (*node_id).num_active_children = 0;   //leave out?
      (*node_id).leave_time          = 0;
      (*node_id).num_fathers         = 0;  //leave out?
      (*node_id).num_active_fathers  = 0;  //leave out?
      for (k=0; k < 2*NUM_STRIPES; k++){
           (*node_id).fathers_id[k]   = NULL;
           (*node_id).rate_of_each[k] = 0.0;
           (*node_id).delay_of_each[k] = 0.0;
           (*node_id).stripe_of_each[k] = 0;
      }// for k
      (*node_id).stripe       = 0;
      
      /* bandwidth assignation */
      rnd_value = uniform_0_1(&seed);
      extracted_dwl_bw = base_chunk_vect[1][0];
      extracted_upl_bw = base_chunk_vect[2][0];
      for (j=0; j<num_deltas_pmf-1; j++){
           if (rnd_value > base_chunk_vect[0][j]){
              extracted_dwl_bw = base_chunk_vect[1][j+1];
              extracted_upl_bw = base_chunk_vect[2][j+1];
           }//if
           else {break;}      
      }//for i
      
      (*node_id).class = j;
      (*node_id).dwl_bw = extracted_dwl_bw/bw_normalization;
      (*node_id).upl_bw = extracted_upl_bw/bw_normalization; 
      (*node_id).spare_dwl_bw = extracted_dwl_bw/bw_normalization;
      (*node_id).spare_upl_bw = extracted_upl_bw/bw_normalization;

      /* populate the neighbor set, up to max_num_neighbors/2;
         start with a random contact_point and borrow some neighbor from the contact,
         reiterating until you have completed the list */
      node_num_neighbors = 0;   
      contact_index = (int)uniform(0.0, (double)num_active_nodes, &seed);
      contact_id = node_list[contact_index];
      visited_nodes = 0;
      while((((*contact_id).leave_time < current_time)||(contact_id == node_id)||((*contact_id).level<1))
             &&(visited_nodes < max_num_peers)) {
             contact_index = (contact_index+1)%max_num_peers; 
             contact_id = node_list[contact_index]; 
             visited_nodes++; 
      }

      if (visited_nodes == max_num_peers){
          num_non_active_nodes++;
          missed_join++;
          num_active_nodes--;
          cumulative_arrived--;
          /* if the node arrives and is not able to attach, wait for the usual 10 chunks */
          new_arrival_time = current_time + negexp(10/(stripe_threshold*stripe_rate), &seed);
          schedule(GENERATION, new_arrival_time, NULL, NULL, 0);
      }
      else{
      
       visited_nodes = 0;
       while((node_num_neighbors < max_num_neighbors/2)&&(visited_nodes < max_num_peers)) {
   
            /* check if the contact is already your neighbor */
            found = FALSE;
            for (i=0; i<node_num_neighbors; i++){
                 if (((*node_id).neighbors[i] == contact_id)||
                      (node_id == contact_id) ) {
                    found = TRUE;
                    break;
                 }       
            }
           
            num_neighbors = (*contact_id).num_neighbors;
            if ((num_neighbors < max_num_neighbors)&&(!found)) {
                (*contact_id).neighbors[num_neighbors] = node_id;
                (*contact_id).num_neighbors++;
                (*node_id).neighbors[node_num_neighbors] = contact_id;
                (*node_id).num_neighbors++;
                node_num_neighbors++;
            }
            neighbor_index = (int)uniform(0.0, (double)num_neighbors, &seed); 
            contact_id = (*contact_id).neighbors[neighbor_index];
            visited_nodes++; 
      }
      node_num_neighbors = (*node_id).num_neighbors; //probably superflous passage, since the variable is updated
      shuffle(node_num_neighbors, (*node_id).neighbors, &seed);
 
      /* finally, schedule the join process */
      schedule(JOIN, current_time, node_id, node_id, 0);

     } 
   }/* if (found == TRUE) */     

}



/*
**  Function  : void search_new_father(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void search_new_father(Node *node_id, Node *from_node, int stripe_id) {

int n_fathers, n_active_fathers, n_children, n_active_children, father_stripe, 
    i, l, best_index, node_index, child_n_fathers, child_n_active_fathers, 
    child_stripe, available_stripe;
Node *father_id, *child_id;   
Booltype found, orphan_of_active;
double father_rate, delay, best_delay, available_rate, delay_of_each;

 /* retrieve information about the father that has left */
 if ((*node_id).end_chunk > 0){
     
     /* search the father that has left among ALL your fathers*/
     found = FALSE;      
     for (i=0; i < 2*NUM_STRIPES; i++){
         father_id  =  (*node_id).fathers_id[i];
         if (father_id == from_node){
             found = TRUE;      
             break;   
         }
     }
    
 /*    #ifdef DEBUG
     if (found==FALSE) {
         printf("n_fathers: %d, found: %d, iterat. %d time: %f node %d\n",
                 n_fathers, found, iterations, current_time, (*node_id).index);
     }           
     #endif

     father_rate      = (*node_id).rate_of_each[i];
     father_stripe    = (*node_id).stripe_of_each[i];
*/
     
     father_rate      = stripe_rate;
     father_stripe    = stripe_id;
     
     n_fathers        = (*node_id).num_fathers;
     n_active_fathers = (*node_id).num_active_fathers;

     /* search among standby fathers iff the node doesn't have enough fathers AND
       has some stanby fathers */
     orphan_of_active = FALSE;
     if (((*node_id).rate <= (stripe_threshold*stripe_rate*0.99))&&(n_active_fathers < n_fathers))
         orphan_of_active = TRUE;
 }
 
 if (((*node_id).end_chunk > 0)&&((*node_id).leave_time != ERROR)){

    /* if orphan_of _active = TRUE, look at your non_active nodes the best delay and make it active; 
       if the new active node has greater delay, register this new delay    */     
    
    if (orphan_of_active){
        n_fathers        = (*node_id).num_fathers;
        n_active_fathers = (*node_id).num_active_fathers;
        best_index = n_fathers;  
        best_delay = HUGE_VAL; 
              
        for (i=n_active_fathers; i < n_fathers; i++){
             father_id        =  (*node_id).fathers_id[i];
             available_rate   =  MIN((*node_id).rate_of_each[i], (*father_id).spare_upl_bw);
             if ((*node_id).leave_time < current_time)
                available_rate = 0;
             delay            =  (*node_id).delay_of_each[i];
             if ((delay<best_delay)&&(available_rate>=father_rate) ){
                 best_index = i;
                 best_delay = delay;
             }
        }
        if (best_index < n_fathers){
            orphan_of_active = FALSE;    
            cumulative_solved++;        
            /* change a father from non-active to active */
            swap_ptr((*node_id).fathers_id, best_index, n_active_fathers);
            swap((*node_id).rate_of_each,   best_index, n_active_fathers);
            swap((*node_id).delay_of_each,  best_index, n_active_fathers);
            swap_int((*node_id).stripe_of_each,   best_index, n_active_fathers);
            (*node_id).num_active_fathers++;
            n_active_fathers++;
            (*node_id).rate         += (*node_id).rate_of_each[n_active_fathers-1];
            (*node_id).spare_dwl_bw -= (*node_id).rate_of_each[n_active_fathers-1];
            /* tell to father it has a new active children */
            father_id         =  (*node_id).fathers_id[n_active_fathers-1];
            n_children        =  (*father_id).num_children;
            n_active_children =  (*father_id).num_active_children;
            found = FALSE;      
            for (j=n_active_children; j < n_children; j++){
                 child_id = (*father_id).neighbors[j];
                 if (child_id == node_id){
                     found = TRUE;      
                     break;
                 }     
            }

            #ifdef DEBUG
            if ((found==FALSE)) {
                printf("found4: %d, iterat. %d time: %f node %d\n", 
                        found, iterations, current_time, (*node_id).index);
            }           
            #endif

            swap_ptr((*father_id).neighbors, j, n_active_children);
            (*father_id).num_active_children++;
            (*father_id).spare_upl_bw -= (*node_id).rate_of_each[n_active_fathers-1];
            /* change the stripe identifier */
            (*node_id).last_change = current_time;
            (*node_id).stripe = 0;
            for (j=0; j < n_active_fathers; j++){
                (*node_id).stripe += ((*node_id).stripe_of_each[j] << half_size_int);
                (*node_id).stripe += (*node_id).stripe_of_each[j];
            }
            for (j=n_active_fathers; j < n_fathers; j++){
                (*node_id).stripe += (*node_id).stripe_of_each[j];
            }
            /* check if the delay is changed */
            best_delay = 0;
            for (j=0; j < n_active_fathers; j++){
                  best_delay = MAX(best_delay, (*node_id).delay_of_each[j]);
            }
            if (best_delay != (*node_id).end_chunk) {
              /* the best candidate has a different delay: the node is forced to change 
                 its delay and notify it to its children */
               node_index = (*node_id).index;   
               stat_delay[X_CONNECT][node_index] = best_delay;
               stat_delay[DELT_POST][node_index] = best_delay - (*node_id).end_chunk;
               (*node_id).end_chunk = best_delay;
               (*node_id).last_change = current_time;
            }
       }
    }//if (orphan_of_active)
       
    /* tell to children that are downloading the stripe that has left that they are orphan;
       this must be done if you have at least an active father, otherwise a "leave" event
       will be scheduled and children will be marked as orphan there */
    if((*node_id).num_active_fathers > 0) {
       n_children        =  (*node_id).num_children;
       n_active_children =  (*node_id).num_active_children;
       for (j=n_children-1; j >= 0; j--){
            child_id  = (*node_id).neighbors[j];
            child_n_fathers        = (*child_id).num_fathers;
            child_n_active_fathers = (*child_id).num_active_fathers;
            found = FALSE;      
            for (k=0; k < child_n_fathers; k++){
                 if ((*child_id).fathers_id[k] == node_id){
                      found = TRUE;      
                      break;   
                 } 
            }
            
            #ifdef DEBUG
            if ((child_n_fathers < 1)||(found==FALSE)) {
                printf("n_fathersN: %d, found: %d, iterat. %d time: %f node %d\n",
                          n_fathers, found, iterations, current_time, (*node_id).index);
            }           
            if ( ((j < n_active_children)&&(k > child_n_active_fathers))||
                 ((j > n_active_children)&&(k < child_n_active_fathers))   )    {
                  printf("mismatch2 active-standby, iterat. %d time: %f node %d, child %d\n",
                          iterations, current_time, (*node_id).index, (*child_id).index);
            }           
            #endif


//////////////////////////////////////////////
            child_stripe = 0;
            for (l=0; l < child_n_fathers; l++){
                child_stripe += (*child_id).stripe_of_each[l];
            }

            available_stripe = (*node_id).stripe >> half_size_int;
            available_stripe = child_stripe^(child_stripe|available_stripe);

            if (((*child_id).stripe_of_each[k] == father_stripe)&&(available_stripe > 0)){
               /* change stripe identifier, look at the delay and other things */
               if (k < child_n_active_fathers)  
                   (*child_id).stripe  -= (father_stripe << half_size_int);
               (*child_id).stripe  -=  father_stripe;
               found = FALSE;      
               delay_of_each = 0;      
               available_stripe = choose_rnd_stripe(available_stripe);
               for (l=0; l < (*node_id).num_active_fathers; l++){
                    if ((*node_id).stripe_of_each[l] == available_stripe){
                         delay_of_each = (*node_id).delay_of_each[l];
                         found = TRUE;      
                         break;
                    }
               }
               
               #ifdef DEBUG
               if ((found==FALSE)) {
                   printf("found0c2: %d, iterat. %d time: %f node %d\n", 
                           found, iterations, current_time, (*node_id).index);
               }           
               #endif

               (*child_id).rate_of_each[k]   = (*node_id).rate_of_each[l];
               (*child_id).delay_of_each[k]  = delay_of_each + 
                                               1/(stripe_threshold*((*child_id).rate_of_each[k]));
               if ((*child_id).delay_of_each[k] > (*child_id).end_chunk){
                  node_index = (*child_id).index;   
                  stat_delay[X_CONNECT][node_index] = (*child_id).delay_of_each[k];
                  stat_delay[DELT_POST][node_index] = (*child_id).delay_of_each[k] - (*child_id).end_chunk;
                  (*child_id).end_chunk = (*child_id).delay_of_each[k];
                  (*node_id).last_change = current_time;
               }        
               (*child_id).stripe_of_each[k] = available_stripe;
               if (k < child_n_active_fathers)  
                   (*child_id).stripe  += (available_stripe << half_size_int);
               (*child_id).stripe  +=  available_stripe;
            
               /* in any case, send an orphan message to your children */
               delay = 0.0;   
//               delay = negexp((1/(stripe_threshold*stripe_rate)), &seed); 
               delay = uniform(0.0,(1/(stripe_threshold*stripe_rate)), &seed); 
               schedule(ORPHAN, current_time  + delay, child_id, node_id, father_stripe);

            }    
    
            else if ((*child_id).stripe_of_each[k] == father_stripe){ 
/////////////////////////////////                   
//            if ((*child_id).stripe_of_each[k] == father_stripe){ 

               if (k < child_n_active_fathers) {  
                   cumulative_orphan++;
                   swap_ptr((*child_id).fathers_id, k, child_n_active_fathers-1);
                   swap((*child_id).rate_of_each,   k, child_n_active_fathers-1);
                   swap((*child_id).delay_of_each,  k, child_n_active_fathers-1);
                   swap_int((*child_id).stripe_of_each,   k, child_n_active_fathers-1);
                      
                   swap_ptr((*child_id).fathers_id, child_n_active_fathers-1, child_n_fathers-1);
                   swap((*child_id).rate_of_each,   child_n_active_fathers-1, child_n_fathers-1);
                   swap((*child_id).delay_of_each,  child_n_active_fathers-1, child_n_fathers-1);
                   swap_int((*child_id).stripe_of_each,   child_n_active_fathers-1, child_n_fathers-1);
                   (*child_id).num_active_fathers--;
           
                   (*child_id).rate         -= (*child_id).rate_of_each[child_n_fathers-1];
                   (*child_id).spare_dwl_bw += (*child_id).rate_of_each[child_n_fathers-1];
                   (*child_id).stripe       -= ((*child_id).stripe_of_each[child_n_fathers-1] << half_size_int);
                   (*child_id).stripe       -=  (*child_id).stripe_of_each[child_n_fathers-1];
                       
                   swap_ptr((*node_id).neighbors, j, n_active_children-1);
                   swap_ptr((*node_id).neighbors, n_active_children-1, n_children-1);
                   (*node_id).num_children--;
                   (*node_id).num_active_children--;
                   n_children--;
                   n_active_children--;
                   (*node_id).spare_upl_bw += (*child_id).rate_of_each[child_n_fathers-1];
               }           
               else {
                   swap_ptr((*child_id).fathers_id, k, child_n_fathers-1);
                   swap((*child_id).rate_of_each,   k, child_n_fathers-1);
                   swap((*child_id).delay_of_each,  k, child_n_fathers-1);
                   swap_int((*child_id).stripe_of_each,   k, child_n_fathers-1);
                   
                   (*child_id).stripe       -=  (*child_id).stripe_of_each[child_n_fathers-1];
                       
                   swap_ptr((*node_id).neighbors, j, n_children-1);
                   (*node_id).num_children--;
                   n_children--;
               }
               (*child_id).num_fathers--;

               delay = 0.0;   
//               delay = negexp((1/(stripe_threshold*stripe_rate)), &seed); 
               delay = uniform(0.0,(1/(stripe_threshold*stripe_rate)), &seed); 
               schedule(ORPHAN, current_time  + delay, child_id, node_id, 
                                                      (*child_id).stripe_of_each[child_n_fathers-1]);
           }// if ((*child_id).stripe_of_each[k] == father_stripe)
        
       }// for (j=n_children-1; j >= 0; j--)
    }// if((*node_id).num_active_fathers > 0)
 
    /* launch the new "update indegree" event*/
    new_arrival_time = current_time + negexp(1/(stripe_threshold*stripe_rate), &seed);
    if ((drastic_change)&&(n_active_fathers < threshold_for_fathers)&&(new_arrival_time < (*node_id).leave_time))
        schedule(CHECK_INDEGREE, new_arrival_time, node_id, node_id, 0);
    
    /* finally, the last case must be managed: the node is still
       orphan of the active node, and, if it has no father, he must leave */
    if((*node_id).num_active_fathers==0) {
       if ((*node_id).leave_time < HUGE_VAL) 
           (*node_id).last_change = (*node_id).leave_time;
       (*node_id).leave_time = ERROR;
       delay = 0.0;
//       delay = negexp((1.0/streaming_rate), &seed);   
       schedule(LEAVE, current_time + delay, node_id, node_id, 0);
    }

    
    /* update temporal statistics for the number of fathers */
    if (fathers_t_index[(*node_id).class] == (*node_id).index) {
          for (i = fathers_t_point[(*node_id).class]; i < current_time; i++){
              fathers_t_run[(*node_id).class][i] = fathers_t_run[(*node_id).class][fathers_t_point[(*node_id).class]];
          }
          fathers_t_run[(*node_id).class][(int)current_time] = (*node_id).num_fathers;
          fathers_t_point[(*node_id).class] = (int)current_time;
    }  

    /* for security, recompute the stripe id */
    n_active_fathers = (*node_id).num_active_fathers;
    n_fathers        = (*node_id).num_fathers;
    (*node_id).stripe = 0;
    if ((*node_id).last_change < current_time)
        (*node_id).last_change = current_time;
    for (j=0; j < n_active_fathers; j++){
        (*node_id).stripe += ((*node_id).stripe_of_each[j] << half_size_int);
        (*node_id).stripe += (*node_id).stripe_of_each[j];
    }
    for (j=n_active_fathers; j < n_fathers; j++){
        (*node_id).stripe += (*node_id).stripe_of_each[j];
    }
    
    
  }// if ((node not left or in error)
}


/*
**  Function  : void search_new_father(int);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
int choose_rnd_stripe(int input_stripe) {


int num_stripes, rnd_int, j, output_stripe;    
            
    num_stripes = 0;
    for (j=0; j < half_size_int; j++){
          if ((input_stripe&((int)(pow(2,j)))) > 0)
              num_stripes++;
         }
    rnd_int = (int)uniform(0.0, (double)num_stripes, &seed);
    j=0;
    while(rnd_int>=0){
          if ((input_stripe&((int)(pow(2,j)))) > 0)
               rnd_int--;
               j++;
          }
    output_stripe = pow(2,j-1);
   
   return(output_stripe);
}   


/*
**  Function  : vcreate_adj_matrix(int, int, int **);
**  Return    : none
**  Remarks   : this funtion ...
**
*/
void create_adj_matrix(int num_nodes, int num_neighbors, int **input_adj_matrix){

int node_i, neighbor_k, *actual_neighbors, *reference_vector;

 /* allocate variuos structures:
    - the vector that memorize the number of neighbors for each node
    - the vector that is shuffled */
  actual_neighbors = calloc(num_nodes, sizeof(int));
  reference_vector = calloc(num_nodes, sizeof(int));
    
  /* analyze each node */
  for (node_i=0; node_i < num_nodes; node_i++) {
     neighbor_k = actual_neighbors[node_i];
  
     if ((num_neighbors - neighbor_k > 1) && 
         (num_nodes - node_i > num_neighbors - neighbor_k)){ 
        /* populate the reference vector and shuffle it*/
        reference_vector[0] = 0;
        for (i=1; i < num_neighbors - neighbor_k; i++)
            reference_vector[i] = 1;
        for (i=num_neighbors - neighbor_k; i < num_nodes - node_i; i++)
            reference_vector[i] = 0;
 
        shuffle_int(num_nodes - node_i, reference_vector, &seed);

        /* populate the adjacency matrix */
        for (i=1; i < num_nodes - node_i; i++) {
            if ((reference_vector[i] == 1)&&(actual_neighbors[i + node_i]<num_neighbors)) {
               input_adj_matrix[node_i][actual_neighbors[node_i]] = i + node_i;
               actual_neighbors[node_i]++;
        
               input_adj_matrix[i + node_i][actual_neighbors[i + node_i]] = node_i;
               actual_neighbors[i + node_i]++;
            }//if
        }//for
    }//if
  }//for
 
  /* space deallocation */   
//  free(actual_neighbors);
//  free(reference_vector);

}  
    











